@@ -1,6 +1,7 @@
 /*
     FUSE Filesystem for libgphoto2.
     Copyright (C) 2005  Philip Langdale <philipl@mail.utexas.edu>
+    Copyright (C) 2007  Marcus Meissner <marcus@jet.franken.de>
 
     This program can be distributed under the terms of the GNU GPL.
     See the file COPYING.
@@ -12,17 +13,18 @@
 
 #include <fuse.h>
 
-#include <gphoto2.h>
+#include <gphoto2/gphoto2.h>
 
 #include <glib.h>
 #include <glib/gi18n.h>
+#include <glib/gprintf.h>
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 #include <fcntl.h>
-
+#include <sys/time.h>
 
 /*
  * Static variables set by command line arguments.
@@ -33,6 +35,8 @@ static gchar *sUsbid = NULL;
 static gint sSpeed = 0;
 static gboolean sHelp;
 
+static struct timeval glob_tv_zero;
+
 /*
  * The OpenFile struct encapsulates a CameraFile and an open count.
  * This allows us to track multiple overlapping open/release calls
@@ -41,7 +45,13 @@ static gboolean sHelp;
 
 struct OpenFile {
    CameraFile *file;
-   guint count;
+   unsigned long count;
+
+   void *buf;
+   unsigned long size;
+   int writing;
+   gchar *destdir;
+   gchar *destname;
 };
 typedef struct OpenFile OpenFile;
 
@@ -122,11 +132,13 @@ struct GPCtx {
    Camera *camera;
    GPContext *context;
    CameraAbilitiesList *abilities;
+   int debug_func_id;
 
    gchar *directory;
    GHashTable *files;
    GHashTable *dirs;
    GHashTable *reads;
+   GHashTable *writes;
 };
 typedef struct GPCtx GPCtx;
 
@@ -161,7 +173,7 @@ gphotofs_readdir(const char *path,
       gchar *key;
 
       stbuf = g_new0(struct stat, 1);
-      stbuf->st_mode = S_IFDIR | 0555;
+      stbuf->st_mode = S_IFDIR | 0755;
       /* This is not a correct number in general. */
       stbuf->st_nlink = 2;
       stbuf->st_uid = getuid();
@@ -199,7 +211,7 @@ gphotofs_readdir(const char *path,
       }
 
       stbuf = g_new0(struct stat, 1);
-      stbuf->st_mode = S_IFREG | 0444;
+      stbuf->st_mode = S_IFREG | 0644;
       stbuf->st_nlink = 1;
       stbuf->st_uid = getuid();
       stbuf->st_gid = getgid();
@@ -226,11 +238,10 @@ exit:
    goto exit;
 }
 
-static int dummyfiller(void *buf,
-                       const char *name,
-                       const struct stat *stbuf,
-                       off_t off)
-{
+static int
+dummyfiller(void *buf, const char *name,
+            const struct stat *stbuf, off_t off
+) {
    return 0;
 }
 
@@ -241,56 +252,55 @@ gphotofs_getattr(const char *path,
 {
    int ret = 0;
    GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+   struct stat *mystbuf = NULL;
+   gpointer value;
+   guint i;
 
    memset(stbuf, 0, sizeof(struct stat));
-
    if(strcmp(path, "/") == 0) {
-      stbuf->st_mode = S_IFDIR | 0555;
+      stbuf->st_mode = S_IFDIR | 0755;
       stbuf->st_nlink = 2;
       stbuf->st_uid = getuid();
       stbuf->st_gid = getgid();
+      return 0;
+   }
+
+   /*
+    * Due to the libgphoto2 api, the best way of verifying
+    * if a file exists is to iterate the contents of that
+    * directory; so if we don't know about the file already,
+    * turn around and call readdir on the directory and try
+    * again.
+    */
+   for (i = 2; i > 0; i--) {
+      gchar *dir;
+      value = g_hash_table_lookup(p->files, path);
+      if (!value) {
+	 value = g_hash_table_lookup(p->dirs, path);
+      }
+      if (value) {
+	 mystbuf = (struct stat *)value;
+	 break;
+      }
+
+      dir = g_path_get_dirname(path);
+      ret = gphotofs_readdir(dir, NULL, dummyfiller, 0, NULL);
+      g_free(dir);
+      if (ret != 0) {
+	 return ret;
+      }
+   }
+
+   if (mystbuf) {
+      stbuf->st_mode = mystbuf->st_mode;
+      stbuf->st_nlink = mystbuf->st_nlink;
+      stbuf->st_uid = mystbuf->st_uid;
+      stbuf->st_gid = mystbuf->st_gid;
+      stbuf->st_size = mystbuf->st_size;
+      stbuf->st_blocks = mystbuf->st_blocks;
+      stbuf->st_mtime = mystbuf->st_mtime;
    } else {
-      struct stat *mystbuf = NULL;
-      gpointer value;
-      guint i;
-
-      /*
-       * Due to the libgphoto2 api, the best way of verifying
-       * if a file exists is to iterate the contents of that
-       * directory; so if we don't know about the file already,
-       * turn around and call readdir on the directory and try
-       * again.
-       */
-      for (i = 2; i > 0; i--) {
-         gchar *dir;
-         value = g_hash_table_lookup(p->files, path);
-         if (!value) {
-            value = g_hash_table_lookup(p->dirs, path);
-         }
-         if (value) {
-            mystbuf = (struct stat *)value;
-            break;
-         }
-
-         dir = g_path_get_dirname(path);
-         ret = gphotofs_readdir(dir, NULL, dummyfiller, 0, NULL);
-         g_free(dir);
-         if (ret != 0) {
-            return ret;
-         }
-      }
-
-      if (mystbuf) {
-         stbuf->st_mode = mystbuf->st_mode;
-         stbuf->st_nlink = mystbuf->st_nlink;
-         stbuf->st_uid = mystbuf->st_uid;
-         stbuf->st_gid = mystbuf->st_gid;
-         stbuf->st_size = mystbuf->st_size;
-         stbuf->st_blocks = mystbuf->st_blocks;
-         stbuf->st_mtime = mystbuf->st_mtime;
-      } else {
-         ret = -ENOENT;
-      }
+      ret = -ENOENT;
    }
    return ret;
 }
@@ -302,36 +312,61 @@ gphotofs_open(const char *path,
    GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
    OpenFile *openFile;
 
-   if((fi->flags & 3) != O_RDONLY) {
-      return -EACCES;
-   }
 
-   openFile = g_hash_table_lookup(p->reads, path);
-   if (!openFile) {
-      gchar *dir = g_path_get_dirname(path);
-      gchar *file = g_path_get_basename(path);
-      CameraFile *cFile;
-      int ret;
+   if ((fi->flags & 3) == O_RDONLY) {
+      openFile = g_hash_table_lookup(p->reads, path);
+      if (!openFile) {
+	 gchar *dir = g_path_get_dirname(path);
+	 gchar *file = g_path_get_basename(path);
+	 CameraFile *cFile;
+	 int ret;
 
-      gp_file_new(&cFile);
-      ret = gp_camera_file_get(p->camera, dir, file, GP_FILE_TYPE_NORMAL,
-                               cFile, p->context);
-      g_free(file);
-      g_free(dir);
+	 gp_file_new(&cFile);
+	 ret = gp_camera_file_get(p->camera, dir, file, GP_FILE_TYPE_NORMAL,
+				  cFile, p->context);
+	 g_free(file);
+	 g_free(dir);
 
-      if (ret != 0) {
-         return gpresultToErrno(ret);
+	 if (ret != 0) {
+	    gp_file_unref (cFile);
+	    return gpresultToErrno(ret);
+         }
+
+	 openFile = g_new0(OpenFile, 1);
+	 openFile->file = cFile;
+	 openFile->count = 1;
+	 g_hash_table_replace(p->reads, g_strdup(path), openFile);
+      } else {
+	 openFile->count++;
       }
-
-      openFile = g_new0(OpenFile, 1);
-      openFile->file = cFile;
-      openFile->count = 1;
-      g_hash_table_replace(p->reads, g_strdup(path), openFile);
-   } else {
-      openFile->count++;
+      return 0;
    }
+   if ((fi->flags & 3) == O_WRONLY) {
+      openFile = g_hash_table_lookup(p->writes, path);
+      if (!openFile) {
+	 gchar *dir = g_path_get_dirname(path);
+	 gchar *file = g_path_get_basename(path);
 
-   return 0;
+	 openFile = g_new0(OpenFile, 1);
+	 openFile->file = NULL;
+	 openFile->count = 1;
+	 openFile->size = 0;
+	 openFile->writing = 1;
+	 openFile->destdir = g_strdup(dir);
+	 openFile->destname = g_strdup(file);
+
+	 openFile->buf = malloc(1);
+	 if (!openFile->buf) return -1;
+	 g_hash_table_replace(p->writes, g_strdup(path), openFile);
+
+         g_free(dir);
+         g_free(file);
+      } else {
+	 openFile->count++;
+      }
+      return 0;
+   }
+   return -EINVAL;
 }
 
 static int
@@ -366,6 +401,188 @@ gphotofs_read(const char *path,
    return ret;
 }
 
+
+/* ================================================================================== */
+
+
+static int
+gphotofs_write(const char *path, const char *wbuf, size_t size,
+                       off_t offset, struct fuse_file_info *fi)
+{
+   GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+   OpenFile *openFile = g_hash_table_lookup (p->writes, path);
+
+   if (!openFile)
+      return -1;
+   if (offset + size > openFile->size) {
+      openFile->size = offset + size;
+      openFile->buf = realloc (openFile->buf, openFile->size);
+   }
+   memcpy(openFile->buf+offset, wbuf, size);
+   return size;
+}
+
+static int
+gphotofs_mkdir(const char *path, mode_t mode)
+{
+    int ret = 0;
+    GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+    gchar *dir = g_path_get_dirname(path);
+    gchar *file = g_path_get_basename(path);
+
+    ret = gp_camera_folder_make_dir(p->camera, dir, file, p->context);
+    if (ret != 0) {
+       ret = gpresultToErrno(ret);
+    } else {
+       struct stat *stbuf;	
+
+       stbuf = g_new0(struct stat, 1);
+       stbuf->st_mode = S_IFDIR | 0555;
+       /* This is not a correct number in general. */
+       stbuf->st_nlink = 2;
+       stbuf->st_uid = getuid();
+       stbuf->st_gid = getgid();
+       g_hash_table_replace(p->dirs, g_strdup (path), stbuf);
+
+    }
+    g_free(dir);
+    g_free(file);
+
+    return ret;
+}
+
+static int
+gphotofs_rmdir(const char *path)
+{
+    int ret = 0;
+
+    GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+    gchar *dir = g_path_get_dirname(path);
+    gchar *file = g_path_get_basename(path);
+
+    ret = gp_camera_folder_remove_dir(p->camera, dir, file, p->context);
+    if (ret != 0) {
+       ret = gpresultToErrno(ret);
+    } else {
+       g_hash_table_remove(p->dirs, path);
+    }
+    g_free(dir);
+    g_free(file);
+    return ret;
+}
+
+static int
+gphotofs_mknod(const char *path, mode_t mode, dev_t rdev)
+{
+   GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+   gchar *dir = g_path_get_dirname(path);
+   gchar *file = g_path_get_basename(path);
+   char *data;
+   int res;
+   CameraFile *cfile;
+
+   gp_file_new (&cfile);
+   data = malloc(1);
+   data[0] = 'c';
+   res = gp_file_set_data_and_size (cfile, data, 1);
+   if (res < 0) {
+      gp_file_unref (cfile);
+      return -1;
+   }
+   res = gp_file_set_name (cfile, file);
+   if (res < 0) {
+      gp_file_unref (cfile);
+      return -1;
+   }
+   res = gp_camera_folder_put_file (p->camera, dir, cfile,
+				    p->context);
+   gp_file_unref (cfile);
+   g_free(dir);
+   g_free(file);
+   return 0;
+}
+
+
+
+static int gphotofs_flush(const char *path, struct fuse_file_info *fi)
+{
+   GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+   OpenFile *openFile = g_hash_table_lookup(p->writes, path);
+
+   if (!openFile)
+      return 0;
+   if (openFile->writing) {
+      int res;
+      CameraFile *file;
+      char *data;
+      gp_file_new (&file);
+      data = malloc (openFile->size);
+      if (!data)
+	 return -ENOMEM;
+      memcpy (data, openFile->buf, openFile->size);
+      /* The call below takes over responsbility of freeing data. */
+      res = gp_file_set_data_and_size (file, data, openFile->size);
+      if (res < 0) {
+	 gp_file_unref (file);
+	 return -1;
+      }
+      res = gp_file_set_name (file, openFile->destname);
+      if (res < 0) {
+	 gp_file_unref (file);
+	 return -1;
+      }
+      res = gp_camera_file_delete(p->camera, openFile->destdir, openFile->destname, p->context);
+      res = gp_camera_folder_put_file (p->camera, openFile->destdir, file, p->context);
+      if (res < 0)
+	 return -ENOSPC;
+      gp_file_unref (file);
+   }
+   return 0;
+}
+
+static int gphotofs_fsync(const char *path, int isdatasync,
+                       struct fuse_file_info *fi)
+{
+    (void) isdatasync;
+    return gphotofs_flush(path, fi);
+}
+
+
+
+static int gphotofs_chmod(const char *path, mode_t mode)
+{
+    return 0;
+}
+
+static int gphotofs_chown(const char *path, uid_t uid, gid_t gid)
+{
+    return 0;
+}
+
+static int gphotofs_statfs(const char *path, struct statvfs *stvfs)
+{
+   GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
+   CameraStorageInformation *sifs;
+   int res, nrofsifs;
+
+   res = gp_camera_get_storageinfo (p->camera, &sifs, &nrofsifs, p->context);
+   if (res < GP_OK)
+      return gpresultToErrno(res);
+   if (nrofsifs == 0)
+      return -ENOSYS;
+
+   if (nrofsifs == 1) {
+      stvfs->f_bsize = 1024;
+      stvfs->f_blocks = sifs->capacitykbytes;
+      stvfs->f_bfree = sifs->freekbytes;
+      stvfs->f_bavail = sifs->freekbytes;
+      stvfs->f_files = -1;
+      stvfs->f_ffree = -1;
+      stvfs->f_favail = -1;
+   }
+   return 0;
+}
+
 static int
 gphotofs_release(const char *path,
                  struct fuse_file_info *fi)
@@ -373,10 +590,17 @@ gphotofs_release(const char *path,
    GPCtx *p = (GPCtx *)fuse_get_context()->private_data;
    OpenFile *openFile = g_hash_table_lookup(p->reads, path);
 
+   if (!openFile) openFile = g_hash_table_lookup(p->writes, path);
+
    if (openFile) {
       openFile->count--;
       if (openFile->count == 0) {
-         g_hash_table_remove(p->reads, path);
+         if (openFile->writing) {
+	     free (openFile->buf);
+             g_hash_table_remove(p->writes, path);
+         } else  {
+             g_hash_table_remove(p->reads, path);
+         }
       }
    }
 
@@ -404,7 +628,6 @@ gphotofs_unlink(const char *path)
    }
 
    g_hash_table_remove(p->files, path);
-
  exit:
    g_free(dir);
    g_free(file);
@@ -413,15 +636,52 @@ gphotofs_unlink(const char *path)
 
 }
 
+static void
+#ifdef __GNUC__
+                __attribute__((__format__(printf,3,0)))
+#endif
+debug_func (GPLogLevel level, const char *domain, const char *format,
+            va_list args, void *data)
+{
+   struct timeval tv;
+   long sec, usec;
+   FILE *logfile = data;
+
+   gettimeofday (&tv,NULL);
+   sec = tv.tv_sec  - glob_tv_zero.tv_sec;
+   usec = tv.tv_usec - glob_tv_zero.tv_usec;
+   if (usec < 0) {sec--; usec += 1000000L;}
+   fprintf (logfile, "%li.%06li %s(%i): ", sec, usec, domain, level);
+   vfprintf (logfile, format, args);
+   fputc ('\n', logfile);
+   fflush (logfile);
+}
+
 static void *
-gphotofs_init(void)
+gphotofs_init()
 {
    int ret = GP_OK;
    GPCtx *p = g_new0(GPCtx, 1);
 
-   p->context = gp_context_new();
+#if 0 /* enable for debugging */
+   int fd = -1;
+   FILE *f = NULL;
 
-   gp_camera_new(&p->camera);
+   fd = open("/tmp/gpfs.log",O_WRONLY|O_CREAT,0600);
+   if (fd != -1) {
+      f = fdopen(fd,"a");
+      if (f)
+         p->debug_func_id = gp_log_add_func (GP_LOG_ALL, debug_func, (void *) f);
+   }
+   fprintf (f, "log opened on pid %d\n", getpid());
+#endif
+
+   p->context = gp_context_new();
+   gettimeofday (&glob_tv_zero, NULL);
+
+   setlocale (LC_CTYPE,"en_US.UTF-8"); /* for ptp2 driver to convert to utf-8 */
+
+   gp_camera_new (&p->camera);
 
    gp_abilities_list_new(&p->abilities);
    gp_abilities_list_load(p->abilities, p->context);
@@ -523,6 +783,8 @@ gphotofs_init(void)
                                     g_free, g_free);
    p->reads = g_hash_table_new_full(g_str_hash, g_str_equal,
                                     g_free, (GDestroyNotify)freeOpenFile);
+   p->writes = g_hash_table_new_full(g_str_hash, g_str_equal,
+                                    g_free, (GDestroyNotify)freeOpenFile);
 
    return p;
 
@@ -576,6 +838,18 @@ static struct fuse_operations gphotofs_oper = {
     .read	= gphotofs_read,
     .release	= gphotofs_release,
     .unlink	= gphotofs_unlink,
+
+    .write	= gphotofs_write,
+    .mkdir	= gphotofs_mkdir,
+    .rmdir	= gphotofs_rmdir,
+    .mknod	= gphotofs_mknod,
+    .flush	= gphotofs_flush,
+    .fsync	= gphotofs_fsync,
+
+    .chmod	= gphotofs_chmod,
+    .chown	= gphotofs_chown,
+
+    .statfs	= gphotofs_statfs
 };
 
 static GOptionEntry options[] =
@@ -599,13 +873,18 @@ main(int argc,
    g_option_context_parse(context, &argc, &argv, &error);
 
    if (sHelp) {
-      guint i;
-      const char *fusehelp[] = { g_get_prgname(), "-ho", NULL};
+      const char *fusehelp[] = { argv[0], "-ho", NULL};
 
       return fuse_main(2, (char **)fusehelp, &gphotofs_oper);
    } else if (sUsbid) {
       g_fprintf(stderr, "--usbid is not yet implemented\n");
+      return 1;
    } else {
-      return fuse_main(argc, argv, &gphotofs_oper);
+     char **newargv = malloc ( (argc+2)*sizeof(char*));
+     memcpy (newargv+2,argv+1,sizeof(char*)*(argc-1));
+     newargv[0] = argv[0];
+     newargv[1] = "-s"; /* disable multithreading */
+
+     return fuse_main(argc+1, newargv, &gphotofs_oper);
    }
 }
